library numerology;

import 'package:basic_utils/basic_utils.dart';
import 'package:intl/intl.dart';
import 'package:numerology/constants.dart';

/// A Calculator.
class NumerologyCalculator {
  /// Return sum of all number in the [text] if the string is digit.
  ///
  /// Otherwise, return null
  ///
  /// [text] must be a digit string
  int _sumNumText(String text) {
    assert(StringUtils.isDigit(text));
    return text
        .split("")
        .map((e) => int.tryParse(e))
        .reduce((value, element) => value = value! + element!)!;
  }

  /// Returns date in dd-MM-yyy format
  String toDdMnYyyy(DateTime dateTime) {
    final df = DateFormat("dd-MM-yyyy");
    return df.format(dateTime);
  }

  /// Return date in ddMMyyyy format
  String toDmy(DateTime dateTime) {
    final dmyString = toDdMnYyyy(dateTime);
    return dmyString.replaceAll("-", "");
  }

  /// Return [Ruling Number](https://destinybirthnumber.tumblr.com/post/85177033226/how-to-calculate-your-ruling-birth-numbers#:~:text=In%20the%20sixth%20century%20B.C.,46%20become%20Ruling%20Number%2010's.)
  /// for the given birthday.
  ///
  /// Ruling number can be [1,11] and 22
  int rulingNumber(DateTime birthday) {
    final dmyString = toDmy(birthday);
    int sum = _sumNumText(dmyString);
    if (sum == 11) return 11;
    if (sum == 22) return 22;
    if (sum == 33) return 33;
    while (sum > 11) sum = _sumNumText("$sum");
    return sum;
  }

  /// Return current [Universal Year Number](https://www.freepressjournal.in/lifestyle/numerology-predictions-for-2021-find-out-whats-in-the-store-for-you)
  ///
  /// It in the range [1, 9]
  ///
  /// This [dateTime] only use for testing purpose
  int universalYearNumber({DateTime? dateTime}) {
    DateTime calDate = DateTime.now();
    if (dateTime != null) calDate = dateTime;
    final yearString = "${calDate.year}";
    int sum = _sumNumText(yearString);
    while (sum > 9) sum = _sumNumText("$sum");
    return sum;
  }

  /// Return current [Personal Year](https://psychiclibrary.com/personal-year-numerology/) for person who was born in [birthday]
  ///
  /// [currentYear] only use for testing purpose
  ///
  /// It must be in range [1,9]
  int personalYear(DateTime birthday, {DateTime? currentYear}) {
    DateTime calYear = DateTime.now();
    if (currentYear != null) calYear = currentYear;
    final universalYear = universalYearNumber(dateTime: calYear);
    final calString = "${birthday.day}${birthday.month}$universalYear";
    int sum = _sumNumText(calString);
    while (sum > 9) sum = _sumNumText("$sum");
    return sum;
  }

  /// Return current [Personal Month](http://numerology.center/month_day_number.php) for person who was born in [birthday]
  ///
  /// It must be in range [1,9]
  int personalMonth(DateTime currentDate, DateTime birthday) {
    int perYear = personalYear(birthday, currentYear: currentDate);
    final calString = "${currentDate.month}$perYear";
    int sum = _sumNumText(calString);
    while (sum > 9) sum = _sumNumText("$sum");
    return sum;
  }

  /// Return [Personal Day](https://www.numerology.com/articles/about-numerology/calculate-personal-day-number/) [currentDate] of person who was born on [birthday]
  ///
  /// It must be in range [1,9] and {11, 22}
  int personalDay(DateTime currentDate, DateTime birthday) {
    final perMonth = personalMonth(currentDate, birthday);
    final calString = "${currentDate.day}$perMonth";
    int sum = _sumNumText(calString);
    if (sum == 11) return 11;
    if (sum == 22) return 22;
    while (sum > 9) sum = _sumNumText("$sum");
    return sum;
  }

  /// Return [Life Path Number](https://www.numerology.com/articles/your-numerology-chart/life-path-number/)
  /// of personal who was born on [birthday]
  ///
  /// It is in range [1,9] and {11, 22, 33}
  int lifePathNumber(DateTime birthday) {
    final sumDay = sumReduceNumTextWithMaster("${birthday.day}");
    final sumMonth = sumReduceNumTextWithMaster("${birthday.month}");
    final sumYear = sumReduceNumTextWithMaster("${birthday.year}");
    final calString = "$sumDay$sumMonth$sumYear";
    return sumReduceNumTextWithMaster(calString);
  }

  /// Returns [Attitude Number](https://www.astronumero.org/birthdate-numerology.html#vedenjsko-stevilo)
  /// of the person who was born in [birthday]
  ///
  /// It must be in range [1, 9]
  int attitudeNumber(DateTime birthday) {
    final calString = "${birthday.day}${birthday.month}";
    int sum = _sumNumText(calString);
    while (sum > 9) sum = _sumNumText("$sum");
    return sum;
  }

  /// Returns [Destiny Number](https://www.prokerala.com/numerology/destiny-numbers.htm#:~:text=It%20is%20calculated%20by%20adding,%3D1%2C%20D%3D4.)
  /// of person who was born on [birthday]
  ///
  /// It must be in range [1, 9]
  int destinyNumber(DateTime birthday) {
    final calString = "${birthday.day}${birthday.month}${birthday.year}";
    final sum = sumReduceNumText(calString);
    return sum;
  }

  /// Return sum reduce to single digit with specical [Master number](https://www.numerology.com/articles/your-numerology-chart/life-path-number/)
  /// , 11, 22, and 33
  ///
  /// It must be in range [1, 9] and {11, 22, 33}
  ///
  /// [text] must be a digit string
  int sumReduceNumTextWithMaster(String text) {
    assert(StringUtils.isDigit(text));
    int sum = _sumNumText(text);
    if (sum == 11) return 11;
    if (sum == 22) return 22;
    if (sum == 33) return 33;
    while (sum > 9) sum = _sumNumText("$sum");
    return sum;
  }

  /// Return sum reduce to single digit in range [1, 9]
  ///
  /// [text] must be a deigt string
  int sumReduceNumText(String text) {
    assert(StringUtils.isDigit(text));
    int sum = _sumNumText(text);
    while (sum > 9) sum = _sumNumText("$sum");
    return sum;
  }

  /// Return [Expression number](https://www.astronumero.org/name-numerology.html#samoizrazno-stevilo) or can be called Mission or Life Purpose number
  /// of person who was born with [fullBirthName]
  ///
  /// It must be in range [1, 9] and {11, 22}
  int expressionNumber(String fullBirthName) {
    final digitName = nameInDigit(fullBirthName);
    int sum = digitName
        .split(" ")
        .map((e) => sumReduceNumTextWithMaster(e))
        .reduce((value, element) => value += element);
    if (_isMasterNum(sum)) return sum;
    sum = sumReduceNumTextWithMaster("$sum");
    return sum;
  }

  /// Returns [Personality number](https://www.astronumero.org/name-numerology.html#osebnostno-stevilo) or can be called
  /// Impression, Outer Self or Adapted Self number
  ///
  /// It must be in range [1, 9] and {11, 22}
  int personalityNumnber(String fullBirthName) {
    final consotantDigitName = consonantNameDigit(fullBirthName);
    int sum = consotantDigitName
        .split(" ")
        .map((e) => sumReduceNumTextWithMaster(e))
        .reduce((value, element) => value += element);
    if (_isMasterNum(sum)) return sum;
    sum = sumReduceNumTextWithMaster("$sum");
    return sum;
  }

  /// Returns [Soul Urge number](https://www.astronumero.org/name-numerology.html#motivacijsko-stevilo) or can be called
  /// Heart Desire or Inner Self number
  ///
  /// It must be in range [1, 9] and {11, 22}
  int soulUrgeNumber(String fullBirthName) {
    final vowelsDigitName = vowelsNameDigit(fullBirthName);
    int sum = vowelsDigitName
        .split(" ")
        .map((e) => sumReduceNumTextWithMaster(e))
        .reduce((value, element) => value += element);
    if (_isMasterNum(sum)) return sum;
    sum = sumReduceNumTextWithMaster("$sum");
    return sum;
  }

  /// Is [num] is master number {11, 22, 33}
  bool _isMasterNum(int num) => num == 11 || num == 22 || num == 33;

  /// Return a name in digits using [Pythagorean Numerology table](https://www.astronumero.org/numerology-charts.html).
  ///
  /// ![Pytharoean Numerology table](https://www.astronumero.org/images/400xNxpythagorean-numerology-chart.jpg.pagespeed.ic.awBQDboCSx.jpg)
  ///
  /// ex: 'John'-> '1685'
  ///
  /// Space will be ignore
  ///
  /// ex: 'John Lennon' -> '1685 355565'
  String nameInDigit(String fullBirthName) {
    assert(fullBirthName.isNotEmpty);
    final convertedString = fullBirthName
        .toLowerCase()
        .split("")
        .map((e) {
          if (e != " ")
            return pythagoreanNumerologyTable[e];
          else
            return e;
        })
        .toList()
        .join("");
    return convertedString;
  }

  /// Return a name in digits only for consonants (include 'y') using [Pythagorean Numerology table](https://www.astronumero.org/numerology-charts.html).
  ///
  /// ![Pytharoean Numerology table](https://www.astronumero.org/images/400xNxpythagorean-numerology-chart.jpg.pagespeed.ic.awBQDboCSx.jpg)
  ///
  /// ex: 'John'-> '185'
  ///
  /// Space will be ignore
  ///
  /// ex: 'John Lennon' -> '185 3555'
  String consonantNameDigit(String fullBirthName) {
    assert(fullBirthName.isNotEmpty);
    final convertedString = fullBirthName.toLowerCase().split("").map((e) {
      if (consonantsWithY.contains(e))
        return pythagoreanNumerologyTable[e];
      else if (e == " ")
        return e;
      else
        return "";
    }).join("");
    return convertedString;
  }

  /// Return a name in digits only vowels using [Pythagorean Numerology table](https://www.astronumero.org/numerology-charts.html).
  ///
  /// ![Pytharoean Numerology table](https://www.astronumero.org/images/400xNxpythagorean-numerology-chart.jpg.pagespeed.ic.awBQDboCSx.jpg)
  ///
  /// ex: 'John'-> '6'
  ///
  /// Space will be ignore
  ///
  /// ex: 'John Lennon' -> '6 56'
  String vowelsNameDigit(String fullBirthName) {
    assert(fullBirthName.isNotEmpty);
    final convertedString = fullBirthName.toLowerCase().split("").map((e) {
      if (vowels.contains(e))
        return pythagoreanNumerologyTable[e];
      else if (e == " ")
        return e;
      else
        return "";
    }).join("");
    return convertedString;
  }
}
